const { EmbedBuilder, ApplicationCommandOptionType } = require("discord.js");
const { EMBED_COLORS } = require("@root/config");

/**
 * @type {import("@structures/Command")}
 */
module.exports = {
  name: "lyric",
  description: "get lyrics for the current song or a specified song",
  category: "MUSIC",
  botPermissions: ["EmbedLinks"],
  command: {
    enabled: true,
    usage: "[song name]",
  },
  slashCommand: {
    enabled: true,
    options: [
      {
        name: "query",
        description: "song name to search lyrics for",
        type: ApplicationCommandOptionType.String,
        required: false,
      },
    ],
  },

  async messageRun(message, args) {
    const query = args.join(" ");
    const response = await getLyric(message, query);
    await message.safeReply(response);
  },

  async interactionRun(interaction) {
    const query = interaction.options.getString("query");
    const response = await getLyric(interaction, query);
    await interaction.followUp(response);
  },
};

async function getLyric({ client, guild, member }, query) {
  /** @type {import('../../handlers/manager')} */
  const manager = client.musicManager
  const player = manager.getPlayer(guild.id)
  let node = player?.node
  if (!node) [node] = manager.nodeManager.nodes.values()

  let track;

  if (!query) {
    // query not specified - use currently playing music
    track = player?.queue.current;
    if (!track) {
      return "🚫 No music is currently playing!";
    }
  } else {
    // query specified -- search
    const result = await node.search({ query }, member.user);
    if (!result || result.loadType === "error" || result.loadType === "empty") {
      return "Failed to find any tracks for the query.";
    }
    track = result.tracks[0];
  }

  // lavalyrics
  try {
    const lyrics = await node.request(`/lyrics?track=${encodeURIComponent(track.encoded)}&skipTrackSource=${true}`)
    if (lyrics && 'text' in lyrics) return createLyricsEmbed(lyrics, member, track)
    else console.log(`Failed search LavaLyrics track ${track.info.title}`, lyrics)
  } catch (e) {
    console.error(`Error search LavaLyrics track ${track.info.title}`, e)
  }

  // javatimed
  if (node.info.plugins.includes('java-timed-lyrics')) {
    // java-timed (current)
    if (!query) {
      try {
        const lyrics = await getCurrentLyricJavaTimed(node, guild.id)
        if (lyrics && 'text' in lyrics) return createLyricsEmbed(lyrics, member, track)
        else console.log(`Failed search JavaTimed (current) track ${track.info.title}`, lyrics)
      } catch (e) {
        console.error(`Error search JavaTimed (current) track ${track.info.title}`, e)
      }
    }

    // java-timed (search)
    try {
      const lyrics = await searchLyricJavaTimed(node, guild.id)
      if (lyrics.lines) {
        for (const line of lyrics.lines) {
          line.timestamp = line.range.start
          line.duration = line.range.end - line.range.start
        }
      }
      if (lyrics && 'text' in lyrics) return createLyricsEmbed(lyrics, member, track)
      else console.log(`Failed search JavaTimed (search) track ${track.info.title}`, lyrics)
    } catch (e) {
      console.error(`Error search JavaTimed (search) track ${track.info.title}`, e)
    }
  }

  return "No lyrics found"
}

function createLyricsEmbed(lyrics, member, track) {
  const embed = new EmbedBuilder()
    .setColor(EMBED_COLORS.BOT_EMBED)
    .setTitle(`${track.info.author} - ${track.info.title}`)
    .setThumbnail(track.info.artworkUrl)
    .setFooter({ text: `Requested by: ${member.user.displayName} | Source: ${lyrics.provider || lyrics.source || lyrics.sourceName || '-'}` });

  const lines = !lyrics.lines ? lyrics.text ?? 'No lyrics found' : lyrics.lines
    .filter(Boolean)
    .map(line => line.line)
    .join("\n");

  const truncatedLyrics = lines.length > 4096 ? `${lines.slice(0, 4093)}...` : lines;

  embed.setDescription(truncatedLyrics);

  return { embeds: [embed] };
}

function getCurrentLyricJavaTimed(node, guildId) {
  return node.request(`/sessions/${node.sessionId}/players/${guildId}/lyrics`)
}

function searchLyricJavaTimed(node, track, source = "genius") {
  return node.request(`/lyrics/search?query=${encodeURIComponent(track.info.title + " " + track.info.author)}&source=${source}`)
}
